/*
 * gnome-keyring
 *
 * Copyright (C) 2018 Red Hat, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, see
 * <http://www.gnu.org/licenses/>.
 *
 * Author: Daiki Ueno
 */

#ifndef __GKD_LOGIN_PASSWORD_H__
#define __GKD_LOGIN_PASSWORD_H__

#include <gio/gio.h>

#define GKD_TYPE_LOGIN_PASSWORD gkd_login_password_get_type ()
G_DECLARE_FINAL_TYPE (GkdLoginPassword, gkd_login_password, GKD, LOGIN_PASSWORD, GTlsPassword);

gboolean         gkd_login_password_get_login_available
                                            (GkdLoginPassword *self);
void             gkd_login_password_set_store_password
                                            (GkdLoginPassword *self,
                                             gboolean store_password);
gboolean         gkd_login_password_get_store_password
                                            (GkdLoginPassword *self);

#endif	/* __GKD_LOGIN_PASSWORD_H__ */
