/*
 * gnome-keyring
 *
 * Copyright (C) 2010 Stefan Walter
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, see
 * <http://www.gnu.org/licenses/>.
 */

#ifndef __GKM_OIDS_H__
#define __GKM_OIDS_H__

#define GKM_OID_EXTUSAGE_SERVER_AUTH "1.3.6.1.5.5.7.3.1"
#define GKM_OID_EXTUSAGE_CLIENT_AUTH "1.3.6.1.5.5.7.3.2"
#define GKM_OID_EXTUSAGE_CODE_SIGNING "1.3.6.1.5.5.7.3.3"
#define GKM_OID_EXTUSAGE_EMAIL "1.3.6.1.5.5.7.3.4"
#define GKM_OID_EXTUSAGE_IPSEC_ENDPOINT "1.3.6.1.5.5.7.3.5"
#define GKM_OID_EXTUSAGE_IPSEC_TUNNEL "1.3.6.1.5.5.7.3.6"
#define GKM_OID_EXTUSAGE_IPSEC_USER "1.3.6.1.5.5.7.3.7"
#define GKM_OID_EXTUSAGE_TIME_STAMPING "1.3.6.1.5.5.7.3.8"

#endif /* __GKM_OIDS_H__ */
